--[[
Name: HealComm-2.0
Revision: $Rev: 57255 $
Author: xbeeps
Website: http://www.wowace.com/wiki/HealComm-2.0
Documentation: http://www.wowace.com/wiki/HealComm-2.0
SVN: svn://svn.wowace.com/wowace/trunk/HealComm-2.0/
Description: Keeps track of healing spells cast by everyone in a group.
Dependencies: AceLibrary, AceEvent-2.0, Babble-Spell-2.2, Roster-2.1, Gratuity-2.0
License: All Rights Reserved
]]

local MAJOR_VERSION = "HealComm-2.0"
local MINOR_VERSION = "$Revision: 57255 $"

if not AceLibrary then error(MAJOR_VERSION .. " requires AceLibrary.") end
if not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then return end
if not AceLibrary:HasInstance("AceEvent-2.0") then error(MAJOR_VERSION .. " requires AceEvent-2.0") end
if not AceLibrary:HasInstance("Babble-Spell-2.2") then error(MAJOR_VERSION .. " requires Babble-Spell-2.2") end
if not AceLibrary:HasInstance("Roster-2.1") then error(MAJOR_VERSION .. " requires Roster-2.1") end
if not AceLibrary:HasInstance("Gratuity-2.0") then error(MAJOR_VERSION .. " requires Gratuity-2.0") end

local HealComm = {};   

-------------------
-- Use Libraries --
-------------------

AceLibrary("AceEvent-2.0"):embed(HealComm);
local BS = AceLibrary("Babble-Spell-2.2");
local Roster = AceLibrary("Roster-2.1");
local Gratuity = AceLibrary("Gratuity-2.0");

-----------------
-- Static Data --
-----------------

-- Cache of spells and heal sizes
local SpellCache = {};

-- Info about spells being cast by other players
local HealTime = {};
local HealTarget = {};
local HealSize = {};

-- Info about the spell being cast by the player
local CastInfo = {}; -- IsCasting, HealingTargetUnitID, HealingTargetName, Size

-- Latency Measurement
local SentTime = 0;
local Latency = 0;

---------------
-- Utilities --
---------------

local function unitFullName(unit)
    name, realm = UnitName(unit);
    if (realm and realm ~= "") then
        return name .. "-" .. realm;
    else
        return name;
    end
end

-- Use this until Roster will understand Realm names in Battlegrounds
local function getUnitIDFromName(name)
    return Roster:GetUnitIDFromName(select(3,string.find(name, "([^-]+)")));
end

local function commSend()
    SendAddonMessage("HealComm", string.format("%s:%d", CastInfo.HealingTargetName, CastInfo.Size), (select(2, IsInInstance()) == "pvp") and "BATTLEGROUND" or "RAID");
end

local function castInfoReset()
    CastInfo.IsCasting = false;
    CastInfo.HealingTargetUnitID = nil;
    CastInfo.HealingTargetName = nil;
    CastInfo.Size = nil;
end

-- Spellbook Scanner --
-- SpellCache[spell][rank][stat]
-- stat: Mana, Heal
local function getBaseSpellInfo(name)

    -- Check if info is already cached
    if (SpellCache[name]) then
        return SpellCache[name];
    end

    SpellCache[name] = {};

    -- Gather info (only done if not in cache)
    local i = 1;

    while true do

        local spellName, spellRank = GetSpellName(i, BOOKTYPE_SPELL);
        
        if (not spellName) then 
            break 
        end

        if (spellName == name) then
            -- This is the spell we're looking for, gather info

            -- Determine rank
	        _, _, spellRank = string.find(spellRank, "(%d+)");
            spellRank = tonumber(spellRank);
            Gratuity:SetSpell(i, BOOKTYPE_SPELL);
    
            -- Determine mana
            local _, _, Mana = Gratuity:Find("(%d+)", 2, 2, nil, true);
            Mana = tonumber(Mana);
            if not (type(Mana) == "number") then Mana = 0 end

            -- Determine healing
            local _, _, HealMin, HealMax = Gratuity:Find("(%d+) ?[\195\160tobisa到~]+ ?(%d+)", 4, 4, nil, true);
            HealMin,HealMax = tonumber(HealMin),tonumber(HealMax);
            local Heal = (HealMin+HealMax)/2;

            SpellCache[spellName][spellRank] = {Mana = Mana, Heal = Heal};
        end
        i = i + 1;
    end
    return SpellCache[name];
end

-- Detects if a buff is present on the unit and returns the application number
local function detectBuff(unit, name, app)
    local i=1;
    local state, apps;
    while true do
        _, _, state, apps = UnitBuff(unit, i);
        if (not state) then 
            return false;
        end
        if (string.find(state, name) and ((app == apps) or (app == nil))) then 
            return apps;
        end
        i=i+1;
    end
end

local function getSpellBonusHealingPenalty(spellLevel, playerLevel)
    local num = spellLevel + 6;

    if (num > playerLevel) then
        return 1;
    else
        return num / playerLevel;
    end
end

local relicSlotNumber = GetInventorySlotInfo("RangedSlot");
local function getEquippedRelicID()
    local itemLink = GetInventoryItemLink('player', relicSlotNumber);
    if (itemLink) then
        local _, _, itemID = string.find(itemLink, "(%d+):");
        return tonumber(itemID);
    end
end

-----------------------------
-- Healing Data Management --
-----------------------------

local function entryDelete(healerName)
    local targetName = HealTarget[healerName];
    HealTime[healerName] = nil;
    HealTarget[healerName] = nil;
    if (targetName and HealSize[targetName]) then
        HealSize[targetName][healerName] = nil;
    end
end

local function entryUpdate(healerName, targetName, healSize, healTime)
    entryDelete(healerName);
    HealTime[healerName] = healTime;
    HealTarget[healerName] = targetName;
    if (not HealSize[targetName]) then 
        HealSize[targetName] = {}; 
    end
    HealSize[targetName][healerName] = healSize;
end

----------------------
-- Public Functions --
----------------------

--[[

Description: Retrieve info about the incoming heals to a specific
             target for heals that are inbound in the time frame
             between the current time and the time specified as
             the second argument. Example:

             GetIncomingHeal("Kaki", GetTime() + 3)

             Retrieves info about the incoming heals on the player
             Kaki. The inquiry is limited to heals that will land
             on the target within the next 3 seconds.

Input:
    targetName - The name of the unit to retrieve information about
    endTime - the desired end time of the inquiry

Output:
    incomingHeal - The total size of the incoming heals within the time frame
    nextTimeLeft - the time until the next heal will land
    nextSize - the size of the next heal
    nextName - the name of the healer casting the next incoming heal

]]--

function HealComm:GetIncomingHeal(targetName, endTime)
    if (HealSize[targetName]) then
        local now = GetTime();
        local incomingHeal;
        local nextTime, nextSize, nextName;
        for healerName,size in pairs(HealSize[targetName]) do
            local time = HealTime[healerName];
            if (size and time) then
                time = time + Latency;
                if (time < now) then
                    -- Overdue
                    entryDelete(healerName);
                elseif (time < endTime)  then
                    -- Due within time frame
                    incomingHeal = (incomingHeal or 0) + size;
                    if ((not nextTime) or (time < nextTime)) then
                        nextTime = time;
                        nextSize = size;
                        nextName = healerName;
                    end
                end
            end
        end
        if (incomingHeal) then
            return incomingHeal, nextTime - now, nextSize, nextName;
        end
    end
end

--[[

Description: Retrieve info about the healing spell currently
             casting by the player.

Input:
    None

Output:
    isCasting - true if player is casting a healing spell
    healingTargetUnitID - Unit ID of the target being healed
    healingTargetName - Name of the unit being healed
    size - Size of the healing being cast

]]--

function HealComm:GetPlayerCastInfo()
    return CastInfo.IsCasting, CastInfo.HealingTargetUnitID, CastInfo.HealingTargetName, CastInfo.Size;
end

function HealComm:IsCastingHeal(healerName)
    local time = HealTime[healerName];
    if (time) then
        if (GetTime() > (time + Latency)) then
            -- Overdue
            entryDelete(healerName);
            return false;
        else
            -- Still casting
            return true;
        end
    elseif (healerName == unitFullName('player')) then
        return CastInfo.IsCasting;
    end
    return false;
end


--------------------
-- Event Handlers --
--------------------

function HealComm:LEARNED_SPELL_IN_TAB()
    -- Invalidate cached spell data when learning new spells
    SpellCache = {};       
end

function HealComm:UNIT_SPELLCAST_SENT(unit, spellName, spellRank, spellTarget)
    if (unit == 'player') then
        SentTime = GetTime();
        if (self.ClassModule.HealingSpells[spellName]) then
            -- Player is casting a healing spell
            CastInfo.HealingTargetName = spellTarget;
            CastInfo.HealingTargetUnitID = getUnitIDFromName(spellTarget) or (spellTarget == unitFullName("focus") and "focus") or (spellTarget == unitFullName("target") and "target") or (spellTarget == unitFullName("targettarget") and "targettarget");
            if (CastInfo.HealingTargetUnitID) then
                CastInfo.Size = self.ClassModule:GetSpellInfo(spellName, tonumber(spellRank:match("(%d+)")), CastInfo.HealingTargetUnitID).EffectiveHeal or 0;
            else
                castInfoReset();
            end
        else
            castInfoReset();
        end
    end
end

function HealComm:UNIT_SPELLCAST_START(unit)
    if (unit == 'player') then
    
        -- Latency measurement
        local currentLatency = GetTime() - SentTime;
        if (currentLatency > 1) then -- Limit to 1 sec
            currentLatency = 1;
        end
        Latency = 0.5 * Latency + 0.70 * currentLatency

        -- Start of actual casting
        if (CastInfo.HealingTargetUnitID) then
            -- Continuation of healing spell cast
            CastInfo.IsCasting = true;
            self:TriggerEvent("HealComm_PlayerStart");
            commSend();
        end
    end
end

function HealComm:CHAT_MSG_ADDON(prefix, msg, channel, healer)
    -- filter out own messages
    if (healer == unitFullName('player')) then
        return;
    end

    if (prefix == "VisualHeal") then -- BACKWARD COMPATIBILITY. The Prefix VisualHeal will be deprecated
        local unitHealer = getUnitIDFromName(healer);
        local _, _, target, size = string.find(msg, "^%d+:(.+):(%d+):");

        if (target and unitHealer) then
            local unitTarget = getUnitIDFromName(target);
            local _, _, _, _, _, endTime = UnitCastingInfo(unitHealer);

            if (unitTarget and endTime) then
                size = tonumber(size);
                entryUpdate(healer, target, size, endTime / 1000);
                self:TriggerEvent("HealComm_UnitStart", unitHealer, unitTarget);
            end
        end
    elseif (prefix == "HealComm") then
        local unitHealer = getUnitIDFromName(healer);
        local _, _, target, size = string.find(msg, "^(.+):(%d+)");

        if (target and unitHealer) then
            local unitTarget = getUnitIDFromName(target);
            local _, _, _, _, _, endTime = UnitCastingInfo(unitHealer);

            if (unitTarget and endTime) then
                size = tonumber(size);
                entryUpdate(healer, target, size, endTime / 1000);
                self:TriggerEvent("HealComm_UnitStart", unitHealer, unitTarget);
            end
        else -- Forward compatibility with LibHealComm-3.0
            local msgtype = tonumber(msg:sub(1, 3));
            if (not msgtype) then return end  

            if (msgtype == 0) then -- DirectHealStart
                local size = tonumber(msg:sub(4, 8));
                local target = msg:sub(8, -1);

                if (target and unitHealer) then
                    local unitTarget = getUnitIDFromName(target);
                    local _, _, _, _, _, endTime = UnitCastingInfo(unitHealer);

                    if (unitTarget and endTime) then
                        entryUpdate(healer, target, size, endTime / 1000);
                        self:TriggerEvent("HealComm_UnitStart", unitHealer, unitTarget);
                    end
                end
            elseif (msgtype == 1) then -- DirectHealStop
                entryDelete(healer);
            end
        end
    end       
end

function HealComm:UNIT_SPELLCAST_DELAYED(unit)
    if (unit == 'player') then
        if (CastInfo.IsCasting) then
            self:TriggerEvent("HealComm_PlayerUpdate");
        end
    elseif (unit ~= 'target' and unit ~= 'focus') then
        local unitName = unitFullName(unit);
        if (unitName and self:IsCastingHeal(unitName)) then
            local _, _, _, _, _, endTime = UnitCastingInfo(unit);
            if (endTime) then
                HealTime[unitName] = endTime / 1000;
                self:TriggerEvent("HealComm_UnitUpdate", unit);
            end
        end
    end
end

function HealComm:UNIT_SPELLCAST_STOP(unit)
    if (unit == 'player') then
        if (CastInfo.IsCasting) then
            CastInfo.IsCasting = false;            
            self:TriggerEvent("HealComm_PlayerUpdate"); 
        end
    elseif (unit ~= 'target' and unit ~= 'focus') then
        local unitName = unitFullName(unit);
        if (unitName and self:IsCastingHeal(unitName)) then
            entryDelete(unitName);
            self:TriggerEvent("HealComm_UnitUpdate", unit);
        end
    end
end

--------------------
-- Class Specific --
--------------------

-- Druid --

HealComm.Druid = {
    HealingSpells = {
        [BS["Healing Touch"]] = {7, 13, 19 , 25, 31, 37, 43, 49, 55, 59, 61, 67, 70},
        [BS["Regrowth"]] = {17, 23, 29, 35, 41, 47, 53, 59, 65, 70}
    }
};

function HealComm.Druid:GetSpellInfo(name, rank, target)
    local i, powerInfusion;

    -- Get static spell info
    local spellInfo = getBaseSpellInfo(name)[rank];

    if (not spellInfo) then
        return nil;
    end

    -- Get +healing bonus
    local bonus = GetSpellBonusHealing();

    local spellTab = self.HealingSpells[name];

    -- Detect Power Infusion
    if (detectBuff('player', "Spell_Holy_PowerInfusion$")) then
        powerInfusion = 1.2;
    else
        powerInfusion = 1.0;
    end

    -- Gift of Nature - Increases healing by 2% per rank on all spells
    local _, _, _, _, talentGiftOfNature = GetTalentInfo(3, 12);
    local talentGiftOfNature = 2 * talentGiftOfNature / 100 + 1;

    -- Process individual spells
    if (name == BS["Healing Touch"]) then
        local nBonus;

        -- Empowered Healing
        local _, _, _, _, talentEmpoweredTouch = GetTalentInfo(3, 14);

        -- Determine normalisation
        if (rank < 5) then
            if (rank == 1) then
                nBonus = (1.5 / 3.5) * bonus * (10 * talentEmpoweredTouch / 100 + 1) * 0.2875;
            elseif (rank == 2) then
                nBonus = (2.0 / 3.5) * bonus * (10 * talentEmpoweredTouch / 100 + 1) * 0.55;
            elseif (rank == 3) then
                nBonus = (2.5 / 3.5) * bonus * (10 * talentEmpoweredTouch / 100 + 1) * 0.775;
            elseif (rank == 4) then
                nBonus = (3.0 / 3.5) * bonus * (10 * talentEmpoweredTouch / 100 + 1);
            end        
        else
            nBonus = bonus * (10 * talentEmpoweredTouch / 100 + 1);
        end

        spellInfo.EffectiveHeal = talentGiftOfNature * powerInfusion * (spellInfo.Heal + nBonus * getSpellBonusHealingPenalty(spellTab[rank], UnitLevel('player')));
    elseif (name == BS["Regrowth"]) then
        local nBonus;
        if (rank == 1) then
            nBonus = (2.0 / 3.5) * bonus * 0.48 * 0.7;
        elseif (rank == 2) then
            nBonus = (2.0 / 3.5) * bonus * 0.5 * 0.925;
        else
            nBonus = (2.0 / 3.5) * bonus * 0.5;
        end
        spellInfo.EffectiveHeal = talentGiftOfNature * powerInfusion * (spellInfo.Heal + nBonus * getSpellBonusHealingPenalty(spellTab[rank], UnitLevel('player')));
    end

    return spellInfo;
end

-- Paladin --

HealComm.Paladin = {
    HealingSpells = {
    [BS["Holy Light"]] = {5, 13, 21, 29, 37, 45, 53, 59, 61, 69, 70},
    [BS["Flash of Light"]] = {25, 33, 41, 49, 57, 65, 70}
    }
};

HealComm.Paladin.Librams = {
    [25644] = 79, -- Blessed Book of Nagrand
    [23006] = 83, -- Libram of Light
    [23201] = 53  -- Libram of Divinity
}

function HealComm.Paladin.GetBlessingOfLightBonus(unit)
    local i=1;

    while true do
        _, _, icon, apps = UnitBuff(unit, i);
        if (not icon) then 
            return 0, 0;
        end
        if (string.find(icon, "Spell_Holy_PrayerOfHealing02") or string.find(icon, "Spell_Holy_GreaterBlessingofLight")) then 
            Gratuity:SetUnitBuff(unit, i);

            local _, _, bonusHL, bonusFoL = Gratuity:Find(" (%d+) .+ (%d+) ", 2, 2, nil, true);
            return (tonumber(bonusHL) or 0), (tonumber(bonusFoL) or 0);
        end
        i=i+1;
    end

end

function HealComm.Paladin:GetSpellInfo(name, rank, target)
    local i, powerInfusion, divineFavor;

    -- Get static spell info
    local spellInfo = getBaseSpellInfo(name)[rank];

    if (not spellInfo) then
        return nil;
    end

    -- Get +healing bonus
    local bonus = GetSpellBonusHealing();

    local spellTab = self.HealingSpells[name];

    -- Detect Power Infusion
    if (detectBuff('player', "Spell_Holy_PowerInfusion$")) then
        powerInfusion = 1.2;
    else
        powerInfusion = 1.0;
    end

    -- Detect Divine Favor
    if (detectBuff('player', "Spell_Holy_Heal$")) then
        divineFavor = 1.5;
    else
        divineFavor = 1.0;
    end

    -- Detect Blessing of Light and Greater Blessing of Light on target
    local bonusHL, bonusFoL = self.GetBlessingOfLightBonus(target);

    -- Healing Light - Increases healing by 4% per rank on all spells
    local _, _, _, _, talentHealingLight = GetTalentInfo(1, 5);
    local talentHealingLight = 4 * talentHealingLight / 100 + 1;

    -- Process individual spells
    if (name == BS["Flash of Light"]) then
        local libramBonus = self.Librams[getEquippedRelicID()] or 0;
        spellInfo.EffectiveHeal = talentHealingLight * divineFavor * powerInfusion * (spellInfo.Heal + ((1.5 / 3.5) * (bonus + libramBonus + bonusFoL)) * getSpellBonusHealingPenalty(spellTab[rank], UnitLevel('player')));
    elseif (name == BS["Holy Light"]) then
        local nBonus;
        if (rank == 1) then
            nBonus = (2.5 / 3.5) * bonus * 0.2875;
        elseif (rank == 2) then
            nBonus = (2.5 / 3.5) * bonus * 0.475;
        elseif (rank == 3) then
            nBonus = (2.5 / 3.5) * bonus * 0.775;
        else
            nBonus = (2.5 / 3.5) * bonus;
        end
        spellInfo.EffectiveHeal = talentHealingLight * divineFavor * powerInfusion * (spellInfo.Heal + (nBonus + bonusHL) * getSpellBonusHealingPenalty(spellTab[rank], UnitLevel('player')));
    end

    return spellInfo;
end

HealComm.Priest = {
    HealingSpells = {
        [BS["Lesser Heal"]] = {3, 9, 15},
        [BS["Heal"]] = {21, 27, 33, 39},
        [BS["Greater Heal"]] = {45, 51, 57, 59, 62, 67, 70},
        [BS["Flash Heal"]] = {25, 31, 37, 43, 49, 55, 60, 66, 70}
    }
};

function HealComm.Priest:GetSpellInfo(name, rank, target)
    local i, powerInfusion;

    -- Get static spell info 
    local spellInfo = getBaseSpellInfo(name)[rank];

    if (not spellInfo) then
        return nil;
    end

    -- Get +healing bonus
    local bonus = GetSpellBonusHealing();

    local spellTab = self.HealingSpells[name];

    -- Detect Power Infusion
    if (detectBuff('player', "Spell_Holy_PowerInfusion$")) then
        powerInfusion = 1.2;
    else
        powerInfusion = 1.0;
    end

    -- Spiritual Healing - Increases healing by 2% per rank on all spells
    local _, _, _, _, talentSpiritualHealing = GetTalentInfo(2, 16);
    local talentSpiritualHealing = 2 * talentSpiritualHealing / 100 + 1;

    -- Process individual spells
    if (name == BS["Lesser Heal"]) then
        local nBonus;
        if (rank == 1) then
            nBonus = (1.5 / 3.5) * bonus * 0.2875;
        elseif (rank == 2) then
            nBonus = (2.0 / 3.5) * bonus * 0.4;
        elseif (rank == 3) then
            nBonus = (2.5 / 3.5) * bonus * 0.625;
        end
        spellInfo.EffectiveHeal = talentSpiritualHealing * powerInfusion * (spellInfo.Heal + nBonus * getSpellBonusHealingPenalty(spellTab[rank], UnitLevel('player')));
    elseif (name == BS["Heal"]) then
        spellInfo.EffectiveHeal = talentSpiritualHealing * powerInfusion * (spellInfo.Heal + ((3.0 / 3.5) * bonus) * getSpellBonusHealingPenalty(spellTab[rank], UnitLevel('player')));
    elseif (name == BS["Greater Heal"]) then
        -- Empowered Healing
        local _, _, _, _, talentEmpoweredHealing = GetTalentInfo(2, 20);
        local nBonus = (3.0 / 3.5) * bonus * (4 * talentEmpoweredHealing / 100 + 1);
        spellInfo.EffectiveHeal = talentSpiritualHealing * powerInfusion * (spellInfo.Heal + nBonus * getSpellBonusHealingPenalty(spellTab[rank], UnitLevel('player')));
    elseif (name == BS["Flash Heal"]) then
        -- Empowered Healing
        local _, _, _, _, talentEmpoweredHealing = GetTalentInfo(2, 20);
        local nBonus = (1.5 / 3.5) * bonus * (2 * talentEmpoweredHealing / 100 + 1);
        spellInfo.EffectiveHeal = talentSpiritualHealing * powerInfusion * (spellInfo.Heal + nBonus * getSpellBonusHealingPenalty(spellTab[rank], UnitLevel('player')));
    end

    return spellInfo;
end

HealComm.Shaman = {
    HealingSpells = {
        [BS["Lesser Healing Wave"]] = {27, 35, 43, 51, 59, 65, 70},
        [BS["Healing Wave"]] = {5, 11, 17, 23, 31, 39, 47, 55, 59, 62, 69, 70},
        [BS["Chain Heal"]] = {45, 53, 60, 67, 70}
    }
};

HealComm.Shaman.Totems = {
    [25645] = 79, -- Totem of The Plains
    [22396] = 80, -- Totem of Life
    [23200] = 53  -- Totem of Sustaining
}

function HealComm.Shaman:GetSpellInfo(name, rank, target)
    local i, powerInfusion;

    -- Get static spell info
    local spellInfo = getBaseSpellInfo(name)[rank];

    if (not spellInfo) then
        return nil;
    end

    -- Get +healing bonus
    local bonus = GetSpellBonusHealing();

    -- Purification Talent (increases healing by 2% per rank)
    local _, _, _, _, talentPurification = GetTalentInfo(3,15); 
    talentPurification = 2 * talentPurification / 100 + 1;

    -- Detect Power Infusion
    if (detectBuff('player', "Spell_Holy_PowerInfusion$")) then
        powerInfusion = 1.2;
    else
        powerInfusion = 1.0;
    end

    local spellTab = self.HealingSpells[name];

    -- Process individual spells
    if (name == BS["Lesser Healing Wave"]) then
        local totemBonus = self.Totems[getEquippedRelicID()];
        spellInfo.EffectiveHeal = powerInfusion * talentPurification * (spellInfo.Heal + ((1.5 / 3.5) * (bonus + (totemBonus or 0))) * getSpellBonusHealingPenalty(spellTab[rank], UnitLevel('player')));
    elseif (name == BS["Healing Wave"]) then
        local nBonus;

        -- Determine normalisation 
        if (rank < 5) then
            if (rank == 1) then
                nBonus = (1.5 / 3.5) * bonus * 0.2875;
            elseif (rank == 2) then
                nBonus = (2.0 / 3.5) * bonus * 0.475;
            elseif (rank == 3) then
                nBonus = (2.5 / 3.5) * bonus * 0.7;
            elseif (rank == 4) then
                nBonus = (3.0 / 3.5) * bonus * 0.925;
            end
        else
            nBonus = (3.0 / 3.5) * bonus;
        end

        -- Detect healing way on target
        local hwMod = detectBuff(target, "Spell_Nature_HealingWay$");
        if (hwMod) then 
            hwMod = 1.0 + 0.06 * hwMod;
        else 
            hwMod = 1.0;
        end;

        spellInfo.EffectiveHeal = hwMod * powerInfusion * talentPurification * (spellInfo.Heal + nBonus * getSpellBonusHealingPenalty(spellTab[rank], UnitLevel('player'))); 
    elseif (name == BS["Chain Heal"]) then
        -- Improved Chain Heal Talent (increases healing by 10% per rank)
        local _, _, _, _, talentImprovedChainHeal = GetTalentInfo(3,19); 
        talentImprovedChainHeal = 10 * talentImprovedChainHeal / 100 + 1;

        spellInfo.EffectiveHeal = talentImprovedChainHeal * powerInfusion * talentPurification * (spellInfo.Heal + ((2.5 / 3.5) * bonus) * getSpellBonusHealingPenalty(spellTab[rank], UnitLevel('player')));
    end

    return spellInfo;
end

----------------
-- Activation --
----------------

local function activate(self)

    local _, playerClass = UnitClass('player');
    playerClass = string.lower(playerClass);

    if (playerClass == "druid") then
        self.ClassModule = self.Druid;
    elseif (playerClass == "paladin") then
        self.ClassModule = self.Paladin;
    elseif (playerClass == "priest") then
        self.ClassModule = self.Priest;
    elseif (playerClass == "shaman") then
        self.ClassModule = self.Shaman;
    end

    -- Unload unused class modules
    self.Druid = nil;
    self.Paladin = nil;
    self.Priest = nil;
    self.Shaman = nil;

    -- Only listen to these events if player is healing class
    if (self.ClassModule) then
        self:RegisterEvent("UNIT_SPELLCAST_SENT");
        self:RegisterEvent("UNIT_SPELLCAST_START");
    end

    self:RegisterEvent("LEARNED_SPELL_IN_TAB");
    self:RegisterEvent("CHAT_MSG_ADDON");
    self:RegisterEvent("UNIT_SPELLCAST_DELAYED");
    self:RegisterEvent("UNIT_SPELLCAST_STOP");
end

AceLibrary:Register(HealComm, MAJOR_VERSION, MINOR_VERSION, activate);
HealComm = nil;